#!usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.ErrorHandle
    ~~~~~~~~~~~~~~~~~~~~~~

    Implements the error handle support for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

import os

    
def valueError(estring):
    raise ValueError(estring)

def ioError(estring):
    raise IOError(estring)

def indexError(estring):
    raise IndexError(estring)

def attrError(estring):
    raise AttributeError



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.RecManipImpl
    ~~~~~~~~~~~~~~~~~~~~~~~

    Implements the really implementing object that manipulator of records for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

import sys
import os
import hashlib
try:
    import cPickle as Pickle
except ImportError:
    import Pickle
from datetime import date, timedelta
from decimal import Decimal
from collections import deque

from Record import BaseRecord
from ConfigManip import Config
from utils import UnicodeReader, UnicodeWriter, to_date
import ErrorHandle as EH


DEFAULT_PASSWORD = '<!--#/**/;;//PyExpenses-->'

class PwdError(Exception):
    # TODO: maybe need to complete some features.
    def __init__(self):
        super(PwdError, self).__init__()
        
class _RecContainer(object):
    """Data container class inner defined."""
    def __init__(self, time):
        self._date = time
        self._storage = deque()

    def addContent(self, base_rec):
        self._storage.append(base_rec)

    def delElem(self, base_rec):
        """delete the specified element."""
        for elem in self._storage:
            if elem == base_rec:
                self._storage.remove(elem)
                return
        else:
            EH.valueError('data is not in Record Manipulator.')

    def updateElem(self, base_rec, new_rec):
        """update the specified element whth new_rec."""
        for elem in self._storage:
            if elem == base_rec:
                elem.copy(new_rec)
                return
            else:
                EH.valueError('data is not in Record Manipulator.')

class PickleImpl(object):
    """Record manipulator that offer some basic operations to mapulate data,
    like access,modify,retrieve etc.

    @Note: The format of seq -- data structure -- of PickleImpl class:

    _reclist[conta1, conta2, conta3, ... , contaN]  # list
             --+---
               |
               +--> _RecContainer(date, storage)  # class
                                        ---+---
                                           |
                  +------------------------+
                  |
                  +--> _storage[baserec1, baserec2, ... , baserecN]  # deque
                                ---+----
                                   |
                      +------------+
                      |
                      +--> BaseRecord(att1, att2, ... , attN)  # class

    """
    def __init__(self, pwd=DEFAULT_PASSWORD, **kwargs):
        if kwargs.get('test', False) is True:
            # use default setting for test
            self._setdefault()
        else:
            self._load(pwd)
        
        # following attributes is the info of records:
        self.sum_amounts = 0.0
        self.sum_items = 0
        
        self.max_item = 0    
        self.max_item_date = ''    # 最多记录项的一天
        
        self.max_amount = 0.0
        self.max_amount_date = ''

    # def __del__(self):
    #     self._dump()

    def updatePassword(self, oldpwd, newpwd):
        m = hashlib.sha256()
        m.update(oldpwd)
        if m.digest() != self._pwd:
            raise PwdError

        m = hashlib.sha256()
        m.update(newpwd)
        self._pwd = m.digest()

    def addItem(self, time, base_rec):
        """add a basic record to storage."""
        if not self.isExistence(time):
            item = _RecContainer(time)
            index = self._index(time)
            self._reclist.insert(index, item)
            self._len += 1

        self._updateStatus_add(base_rec)
        self._insert(time, base_rec)

    def delItem(self, date, base_rec):
        """delete a specified record.
        @note: if date not exist,raise a exception.
        """
        if not self.isExistence(date):
            EH.valueError('{0} is not in Record Manipulator.'.format(date))
        
        index = self._index(date)
        self._reclist[index].delElem(base_rec)

        if not len(self._reclist[index]._storage):
            # the element of _reclist has not BaseRecord
            del self._reclist[index]

    def updateItem(self, date, base_rec, new_rec):
        if not self.isExistence(date):
            EH.valueError('{0} is not in Record Manipulator.'.format(date))

        index = self._index(date)
        self._reclist[index].updateElem(base_rec, new_rec)

    def getAll(self):
        ret = {}
        for elem in self._reclist:
            ret[elem._date] = elem._storage
        return ret

    def getInfo(self):
        pass

    def date_range(self, begin, end):
        """Return a range of records in specified period.
        @note: 
        """
        if begin > end:
            EH.indexError(
                'Begin time {0} greater than end time {1}'.format(begin, end)
            )
        bindex = self._index(begin)
        eindex = self._index(end)

        ret = {}
        for elem in self._reclist[bindex:eindex]:
            ret[elem._date] = elem._storage
        return ret

    def clear(self):
        """clear all the data in data manipulator."""
        for elem in self._reclist[:]:
            for ite in xrange(len(elem._storage)):
                del elem._storage[0]
            del elem._date
            
        for ite in xrange(len(self._reclist)):
            del self._reclist[0]
        self._setdefault()
        # save the empty status
        self.save()

    def save(self):
        self._dump()
        
    def importRecord(self, filename):
        """import a backup file that was implemented by CSV format from 
        specify file, note that it will override original data."""
        self.clear()  # NOTE: maybe should make this operation decided in upper level.
        reader = UnicodeReader(open(filename, 'rb'))
        for row in reader:
            rdate = to_date(row[0])
            rtype = row[2], row[3]
            temprec = BaseRecord(row[1], rtype, row[4], row[5], row[6], row[7])
            self.addItem(rdate, temprec)

    def exportRecord(self, filename):
        """Export textual data records that is implemented by CSV format
        into the specified file.
        """
        writer = UnicodeWriter(open(filename, 'wb'))
        outlist = []
        for elem in self._reclist:
            rdate = elem._date.isoformat()
            for item in elem._storage:
                outrec = [rdate, 
                          str(item.amount).encode('utf-8'),
                          item.type[0],
                          item.type[1],
                          item.payment,
                          item.currency,
                          item.tag,
                          item.comment]
                outlist.append(outrec)
                
        writer.writerows(outlist)

    def _dump(self):
        fpath = os.path.join(Config.getInfo('path'), 'records.db')
        picfile = open(fpath, 'wb')
        Pickle.dump(self._reclist, picfile, 2)
        Pickle.dump(self._pwd, picfile, 2)
        Pickle.dump(self._len, picfile, 2)
        picfile.close()

    def _load(self, pwd):
        fpath = os.path.join(Config.getInfo('path'), 'records.db')
        isfile = os.path.isfile(fpath)
        if not isfile or not os.path.getsize(fpath):
            self._setdefault()
            if not isfile:
                # need to create a new store file of data
                self._dump()
        else:
            picfile = open(fpath)
            self._reclist = Pickle.load(picfile)
            self._pwd = Pickle.load(picfile)
            self._len = Pickle.load(picfile)
            picfile.close()
        
        m = hashlib.sha256()
        m.update(pwd)
        if m.digest() != self._pwd:
            raise PwdError

    def _setdefault(self):
        self._reclist = []
        m = hashlib.sha256()
        # There is a default common pwd digest.
        m.update(DEFAULT_PASSWORD)
        self._pwd = m.digest()
        self._len = 0

    def isExistence(self, dateitem):
        """Judge whether the item exist or not."""
        low = 0
        high = self._len - 1
        while True:
            if low > high:
                return False
            mid = (low + high) / 2
            try:            
                if self._reclist[mid]._date < dateitem:
                    low = mid + 1
                elif self._reclist[mid]._date == dateitem:
                    return True
                elif self._reclist[mid]._date > dateitem:
                    high = mid - 1
            except IndexError:
                print 'len = ', self._len, len(self._reclist)
                print 'low, mid, high = ', low, mid, high
                sys.exit(1)

    def _index(self, date):
        low = 0
        high = self._len - 1
        while True:
            # if can't find element then return the most matching position
            if low > high:
                return low
            mid = (low + high) / 2
            
            if self._reclist[mid]._date < date:
                low = mid + 1
            elif self._reclist[mid]._date == date:
                return mid
            elif self._reclist[mid]._date > date:
                high = mid - 1

    def _updateStatus_add(self, baserec):
        self.sum_amounts += baserec.amount
        self.sum_items += 1

    def _updateStatus_del(self, baserec):
        self.sum_amounts -= baserec.amount
        self.sum_items -= 1

    def _updateStatus_upd(self, baserec, newrec):
        self.sum_items += (baserec.amount - newrec.amount)
                
    def _insert(self, date, record):
        i = self._index(date)
        self._reclist[i].addContent(record)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.ProjectParse
    ~~~~~~~~~~~~~~~~~~~~~~~

    Implement the parsing functions for Projects.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

from datetime import date, timedelta

from utils import to_date


PROJECT_TARGET = 'The amount of project target is {0}.'
PROJECT_START_DATE = 'The project was start at {0}.'
REST_OF_GOAL = 'For accomplish project target, remain {0}.'
AVERAGE_OF_REC = 'The average of each amount of registering record is {0}.'
PAST_DATES = 'From project start, had {0} days was past.'
REMAIN_DATES = 'To project end, had {0} days is remain.'
HAS_ACCOMPLISH = 'The project was accomplished at {0}.'

HAS_QUARTER = 'The quarter target amount of project had satified at {0}.'
HAS_HALF = 'The half target amount of project had satified at {0}.'
HAS_THREEQUARTERS = 'The three-quarter target amount of project had satified at {0}.'
IS_RECURSION = 'Project recursion is enable, recursive period is {0} days.'

    
def consumingProjectParse(data):
    retseq = []
    statdict = data['p_statdict']

    restOfGoal = data['p_target_amount'] - data['p_accumt_amount']
    averageOfRec = data['p_accumt_amount'] / data['p_recs_amount']
    pastDates = (date.today() - to_date(data['p_start_time'])).days
    remainDates = data['p_period'] - pastDates

    # figure out the basic infomation of project.
    templist = [
        PROJECT_TARGET.format(data['p_target_amount']),
        PROJECT_START_DATE.format(data['p_start_time']),
        REST_OF_GOAL.format(restOfGoal),
        AVERAGE_OF_REC.format(averageOfRec),
        PAST_DATES.format(pastDates),
        REMAIN_DATES.format(remainDates)
    ]
    retseq.extend(templist)

    # figure out the milestone of project target.
    _milestone_check(
        data['p_accumt_amount'], data['p_target_amount'], statdict
    )
    # judge situations of project to process 
    if data['p_accomplish_date']:
        retseq.append(
            HAS_ACCOMPLISH.format(data['p_accomplish_date'])
        )
    elif statdict['hasThreeQuarters']:
        retseq.append(
            HAS_THREEQUARTERS.format(statdict['DateOfThreeQuarters'])
        )
    elif statdict['hasHalf']:
        retseq.append(
            HAS_HALF.format(statdict['DateOfHalf'])
        )
    elif statdict['hasQuarter']:
        retseq.append(
            HAS_QUARTER.format(statdict['DateOfQuarter'])
        )

    # if project recursion is enable.
    if data['p_recursion']:
        retseq.append(
            IS_RECURSION.format(data['p_period'])
        )

    return retseq

def _milestone_check(accumulated, target, statdict):
    if not statdict['hasQuarter']:
        if accumulated >= (target / 4):
            statdict['hasQuarter'] = True
            statdict['DateOfQuarter'] = date.today().isoformat()
    if not statdict['hasHalf']:
        if accumulated >= (target / 2):
            statdict['hasHalf'] = True
            statdict['DateOfHalf'] = date.today().isoformat()
    if not statdict['hasThreeQuarters']:
        if accumulated >= (target * 3 / 4):
            statdict['hasThreeQuarters'] = True
            statdict['DateOfThreeQuarters'] = date.today().isoformat()



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.RecParser
    ~~~~~~~~~~~~~~~~~~~~

    Implements the class of parsers for records parsing.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.

----------------------------------------------------
THE FORMAT OF PARSING SEQUENCE THAT USE FOR PARSERS:

    Sequence[date1, date2, date3, ... , dateN]  # dict
             --+--
               | (mapping) (date : storage)
               |
               +--> Storage[baserec1, baserec2, ... , baserecN]  # list
                            ---+----
                               |
                       +-------+
                       |
                       +--> BaseRecord(att1, att2, ... , attN)  # class
"""

from decimal import Decimal
from datetime import date, timedelta

import ErrorHandle as EH


def asum(datas):
    """Calculate the sum of amount in sequence.
    """
    amount = 0.0
    # items structure is a tuple such as (date, recseq)
    for key in datas:
        amount += sum(record.amount for record in datas[key])
    #Note: Decimal use quantize() to ensure output precision.
    return Decimal(str(amount)).quantize(Decimal('1.00'))


class Parser(object):
    def __init__(self, parser_type = None):
        self.type = parser_type
        
    def parse(self):
        pass


class MainParser(Parser):
    def __init__(self, recgroup):
        self._records = recgroup            # data sequence
        self._ft_list = []            # filter list
        self._stat_list = []            # statistic list
        self._analy_list = []            # analysis list
        super(MainParser, self).__init__('mainparser')

    def parse(self):
        if not self._stat_list and not self._analy_list:
            # default strategy is use General_Analys to parse the records.
            self._analy_list.append(General_Analys())

        dealed_seq = self._filter()
        analy_result = self._analysis(dealed_seq)
        stat_result = self._statistic(dealed_seq)
        
        return analy_result, stat_result, dealed_seq

    def _filter(self):
        """filter data container."""
        if not self._ft_list:
            return self._records
        
        dealed_seq = self._records.copy()
        for parser in self._ft_list:
            dealed_seq = parser.parse(dealed_seq)
        return dealed_seq

    def _statistic(self, recgroup):
        """statistic data container."""
        retval = []
        for parser in self._stat_list:
            retval.append(parser.parse(recgroup))
        return retval

    def _analysis(self, recgroup):
        """analyze data container."""
        retval = []
        for parser in self._analy_list:
            retval.append(parser.parse(recgroup))
        return retval

    def append(self, parser):
        if parser.type == 'filter':
            self._ft_list.append(parser)
        elif parser.type == 'statistic':
            self._stat_list.append(parser)
        elif parser.type == 'analysis':
            self._analy_list.append(parser)
        else:
            EH.valueError("{0} is can't recognize.".format(parser.type))

    def extend(self, plist):
        for parser in plist:
            self.append(parser)

    def clear(self):
        self._ft_list = []
        self._stat_list = []
        self._analy_list = []

    def getchild(self):
        pass
    
        
    
class Amount_Stat(Parser):
    """Figure out the amount of money in data flow.
    @note: can specify a few subparser that use to filter some
           situation.
    """
    def __init__(self):
        super(Amount_Stat, self).__init__('statistic')

    def parse(self, recgroup):
        return 'Amount', asum(recgroup)


class Count_Stat(Parser):
    """Figure out the amount of records in data flow.
    @note: 是否应该 可选择是否返回Recs的具体项，待考虑。
    """
    def __init__(self):
        super(Count_Stat, self).__init__('statistic')

    def parse(self, recgroup):
        """Count the quantity of elements in sequence.
        """
        csum = sum(len(item) for item in recgroup.values())
        return 'Count', csum

    
class Type_Filter(Parser):
    """用于extract指定Type的Recores数据。"""
    def __init__(self, Optcode, AttrType='type'):
        self.optcode = Optcode
        self.atype = AttrType
        self._test()
        super(Type_Filter, self).__init__('filter')

    def parse(self, recgroup):
        if self.atype is 'type' and not isinstance(self.optcode, tuple):
            # BaseRecord.type is constructed to (maintype, subtype), is a
            # 2-tuple, this situation is just filter maintype of type.
            return self._type_filter(recgroup, ismaintype=True)
        else:
            return self._type_filter(recgroup)

    def _type_filter(self, recgroup, ismaintype=False):
        retval = {}
        for kdate, records in recgroup.items():
            for record in records:
                rtype = self._gettype(record, ismaintype)
                if rtype == self.optcode:
                    if not retval.has_key(kdate):
                        retval[kdate] = []
                    retval[kdate].append(record)
        return retval

    def _gettype(self, record, isMaintype):
        if isMaintype:
            return getattr(record, 'type')[0]
        else:
            return getattr(record, self.atype, None)

    def _test(self):
        pass


class Money_Filter(Parser):
    """用于ectract指定金额范围内的Recores数据。"""
    def __init__(self, range1, range2 = None):
        if range2:
            self.start = float(range1)
            self.stop = float(range2)
        else:
            self.start = 0.0
            self.stop = float(range1)

        self._test()
        super(Money_Filter, self).__init__('filter')

    def parse(self, recgroup):
        retval = {}
        for kdate, records in recgroup.items():
            for record in records:
                if self.start <= record.amount < self.stop:
                    if not retval.has_key(kdate):
                        retval[kdate] = []
                    retval[kdate].append(record)
        return retval

    def _test(self):
        if self.start < 0 or self.stop < 0 or \
        self.stop < self.start:
            EH.valueError("start number({0}) and stop number({1}) has" \
                              "value problem.".format(self.start, self.stop))

    
class PT_Filter(Parser):
    pass


class DaysInWeek_Filter(Parser):
    """Filter the days in weekday that specified by arglist.

    @note: about the days number, where Monday is 0 and Sunday is 6.
    """
    def __init__(self, arglist):
        self.weekdays = self._test_and_correct(arglist)
        super(DaysInWeek_Filter, self).__init__('filter')

    def parse(self, recgroup):
        retval = {}
        for kdate, records in recgroup.items():
            if kdate.weekday() in self.weekdays:
                retval[kdate] = records
        return retval

    def _test_and_correct(self, daylist):
        for i in xrange(len(daylist)):
            daylist[i] = int(daylist[i])
            
        for item in daylist:
            if not 0 <= item <= 6:
                EH.valueError(
                    '{0} is not a correct number in day of weekdays.'.format(item)
                )
        return daylist


class DaysInMonth_Filter(Parser):
    """Filter the days in day of month that specified by arglist.
    """
    def __init__(self, arglist):
        self.days = self._test_and_correct(arglist)
        super(DaysInMonth_Filter, self).__init__('filter')

    def parse(self, recgroup):
        retval = {}
        for kdate, records in recgroup.items():
            if kdate.day in self.days:
                retval[kdate] = records
        return retval

    def _test_and_correct(self, daylist):
        for i in xrange(len(daylist)):
            daylist[i] = int(daylist[i])
            
        for item in daylist:
            if not 1 <= item <= 31:
                EH.valueError(
                    '{0} is not a correct number in day of months.'.format(item)
                )
        return daylist


class MthsInYear_Filter(Parser):
    """Filter the months in year that specified by arglist.
    """
    def __init__(self, arglist):
        self.mths = self._test_and_correct(arglist)
        super(MthsInYear_Filter, self).__init__('filter')

    def parse(self, recgroup):
        retval = {}
        for kdate, records in recgroup.items():
            if kdate.month in self.mths:
                retval[kdate] = records
        return retval

    def _test_and_correct(self, mthlist):
        for i in xrange(len(mthlist)):
            int(mthlist[i])
            
        for item in mthlist:
            if not 1 <= item <= 12:
                EH.valueError('{0} is not a correct number in months.' \
                                  .format(item))
        return mthlist


class General_Analys(Parser):
    """There are provides some general functionality for analysis 
    include figure out amount of different consuming type and the
    percentage of those amount.

    @note: the format of return values is:
                item (type, amount, percentage)
    """
    def __init__(self):
        super(General_Analys, self).__init__('analysis')

    def parse(self, recgroup):
        valdict = {}
        for kdate, records in recgroup.items():
            for record in records:
                # this parser just analysis the maintype of BaseRecord.type
                # so that use record.type[0] to be the key of dict followed.
                key = record.type[0]
                if not valdict.has_key(key):
                    valdict[key] = [0.0, None]

                valdict[key][0] += record.amount

        asum = sum(item[0] for item in valdict.values())
        self._figure_percentage(valdict, asum)

        retval = self._make_list(valdict)
        retval.sort(key=lambda x: x[1])
        return 'Generally Analyse', retval

    def _figure_percentage(self, valdict, asum):
        for key in valdict.keys():
            percent = valdict[key][0] / asum * 100
            valdict[key][1] = '{0:.2f}'.format(percent)  # convert to formated str

    def _make_list(self, valdict):
        """From dictionary struc transform to list struc."""
        retlist = []
        for key in valdict.keys():
            elem = (key,
                    Decimal(str(valdict[key][0])).quantize(Decimal('1.00')),
                    valdict[key][1])
            retlist.append(elem)
        return retlist



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.RecManip
    ~~~~~~~~~~~~~~~~~~~

    Implements the interface's object that manipulator of records for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

from datetime import date, timedelta

from ConfigManip import Config
import RecManipImpl

STORAGE_BACKEND = RecManipImpl.__dict__[Config.getInfo('StorageBackend')]

class RecManip(object):

    def setUp(self, *args, **kwargs):
        """It's the method really initialize the instance. This should be
        called after create a instance, the passing arguments can control
        defferent options.
        """
        self.impl = STORAGE_BACKEND(*args, **kwargs)

    def updatePassword(self, oldpwd, newpwd):
        self.impl.updatePassword(oldpwd, newpwd)

    def cancelPassword(self, oldpwd):
        from RecManipImpl import DEFAULT_PASSWORD
        self.impl.updatePassword(oldpwd, DEFAULT_PASSWORD)

    def addItem(self, rdate, base_rec):
        """add a basic record to storage."""
        self.impl.addItem(rdate, base_rec)

    def delItem(self, rdate, base_rec):
        """delete a specified record.
        @NOTE: if date not exist,raise a exception.
        """
        self.impl.delItem(rdate, base_rec)

    def updateItem(self, rdate, base_rec, new_rec):
        self.impl.updateItem(rdate, base_rec, new_rec)

    def getAll(self):
        return self.impl.getAll()

    def getInfo(self):
        return self.impl.getInfo()

    def date_range(self, begin, end):
        """Return a range of records in specified period.
        @NOTE:
        """
        return self.impl.date_range(begin, end)

    def findDate(self, fdate):
        """find and return the matching record. 
        @NOTE: if date not exist,raise a exception.
        """
        return self.impl.date_range(fdate, fdate + timedelta(1))

    def findDates(self, fdate, nth=0):
        return self.impl.date_range(fdate, fdate + timedelta(nth))

    def clear(self):
        """clear all the data in data manipulator."""
        self.impl.clear()

    def save(self):
        """save all the data in data manipulator to permanent storage."""
        self.impl.save()

    def importRecord(self, filename):
        self.impl.importRecord(filename)

    def exportRecord(self, filename):
        self.impl.exportRecord(filename)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.utils
    ~~~~~~~~~~~~~~~~

    Implements various utilities functions or class.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

import csv
import cStringIO
import codecs
from datetime import date

import ErrorHandle as EH
from Record import BaseRecord


def datecheck(date):
    """检查date的格式是否符合ISO格式 'yyyy-mm-dd' """
    if date[4] == '-' and date[7] == '-':
        return True
    return False

def to_date(da):
    """假设da的格式为'yyyy-mm-dd'."""
    if isinstance(da, date):
        return da
    elif isinstance(da, (str, unicode)):
        if datecheck(da):
            y = int(da[:4])
            m = int(da[5:7])
            d = int(da[8:])
            return date(y, m, d)
    else:
        EH.valueError('date ({0}) format error.'.format(str(da)))

def buf2record(buf):
    rdate = to_date(buf[0])
    rtype = buf[2], buf[3]
    rec = BaseRecord(buf[1], rtype, buf[4], buf[5], buf[6], buf[7])
    return rdate, rec

def record2buf(rdate, rec):
    return [
        rdate.isoformat(),
        rec.amount,
        rec.type[0],
        rec.type[1],
        rec.payment,
        rec.currency,
        rec.tag,
        rec.comment
    ]


class UTF8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")

class UnicodeReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self

class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.ConfigManip
    ~~~~~~~~~~~~~~~~~~~~~~

    Implements the configuration related objects.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

import json
import os
from os.path import abspath, join, dirname

import ErrorHandle as EH

def _importObj(filename):
    path = join(abspath(dirname(__file__)), filename)
    
    if not os.path.isfile(path):
        raise IOError
    if not os.path.getsize(path):
        # the size of file is zero that contain nothing
        with open(path, 'w') as tempfile:
            print '\n** ', path, '\n'
            json.dump(dict(), tempfile)   # create a empty json dict

    return json.load(open(path))

def _exportObj(obj, filename, path=None):
    if path:
        fpath = os.path.join(path, filename)
    else:
        fpath = os.path.join(Config.getInfo('path'), filename)
    with open(fpath, 'w') as jfile:
        json.dump(obj, jfile, sort_keys=True, indent=4)

def _updOrder(item, seq):
    index = seq.index(item)
    if index == 0:
        # don't need to change order because item is the first element of seq
        return seq
    else:
        return seq[index:index+1] + seq[:index] + seq[index+1:]

def _addTypeOpt(obj, mtype, subtype):
    """If don't understand the logic of this function, could look up
    the config.json file for clear mind.
    """
    if obj['Type']['types'].has_key(mtype):
        if subtype not in obj['Type']['types'][mtype]:
            obj['Type']['types'][mtype].append(subtype)
    else:
        # Note: before assignment subtype should convert to a list firstly.
        # Also note that subtype could't use list() to init because subtype
        # is a str or unicode object.
        obj['Type']['types'][mtype] = [subtype]

def _delTypeOpt(obj, mtype, subtype):
    if obj['Type']['types'].has_key(mtype):
        if subtype in obj['Type']['types'][mtype]:
            obj['Type']['types'][mtype].remove(subtype)
            if not obj['Type']['types'][mtype]:
                del obj['Type']['types'][mtype]
                

class ConfiMeta(type):
    def __new__(cls, name, bases, dict):
        try:            
            dict['objfile'] = 'config.json'
            dict['buffile'] = 'databuf.json'
            dict['obj'] = _importObj(filename=dict['objfile'])
            # A storing object that contain the datas of project's buffer
            # and records buffer by json format.
            dict['bufobj'] = _importObj(filename=dict['buffile'])

            dict['obj']['BaseInfo']['path'] = abspath(dirname(__file__))
        except IOError:
            EH.ioError(
                "file ./confi.json not exist, can't initialise program."
            )
        return super(cls, ConfiMeta).__new__(cls, name, bases, dict)

class Config(object):
    __metaclass__ = ConfiMeta
    
    @classmethod
    def getInfo(cls, option):
        return cls.obj['BaseInfo'][option]
            
    @classmethod
    def setInfo(cls, option, value):
        """Set the given option to the specifit value for BaseInfo section.
        @note: there are not to check the value is valid or not, so need
               to ensure value is valid before invoke this method.
        """
        if not cls.obj['BaseInfo'].has_key(option):
            EH.attrError(
                'Expense info has not "{0}" option.'.format(option)
            )
        cls.obj['BaseInfo'][option] = value
        _exportObj(cls.obj, cls.objfile)

#    @classmethod
#    def secionts(cls):
#        return cls.obj.keys()
            
    @classmethod
    def getOptions(cls, section):
        if section not in ['Type', 'Payment', 'Currency', 'Tag']:
            EH.valueError('section value is invalid.')
        retval = []
        if section == 'Type':
            for t in cls.obj[section]['type_order']:
                retval.append(
                    (t, cls.obj['Type']['types'][t])
                )
        else:
            retval = cls.obj[section]['types']
        return retval

    @classmethod
    def addOption(cls, section, val):
        """If val is not exist then create a new one that suit to the
        situation of section. On the other hand, if val is already exist
        nothing will happends.
        """
        if section not in ['Type', 'Payment', 'Currency', 'Tag']:
            EH.valueError('section value is invalid.')
        if section == 'Type':
            _addTypeOpt(cls.obj, mtype=val[0], subtype=val[1])
        else:
            if val not in cls.obj[section]['types']:
                cls.obj[section]['types'].append(val)
        _exportObj(cls.obj, cls.objfile)

    @classmethod
    def delOption(cls, section, val):
        if section not in ['Type', 'Payment', 'Currency', 'Tag']:
            EH.valueError('section value is invalid.')
        if section == 'Type':
            _delTypeOpt(cls.obj, mtype=val[0], subtype=val[1])
        else:
            if val in cls.obj[section]['types']:
                cls.obj[section]['types'].remove(val)
        _exportObj(cls.obj, cls.objfile)

    @classmethod
    def getDefaults(cls):
        """This method offer the default options of Type, Payment, Currency
        and Tag.
        """
        retval = []
        retval.append(cls.obj['Type']['default'])
        retval.append(cls.obj['Payment']['default'])
        retval.append(cls.obj['Currency']['default'])
        retval.append(cls.obj['Tag']['default'])
        return retval

    @classmethod
    def setDefault(cls, baserec):
        """This method serve for the options include Type, Payment, Currency
        and Tag, provide recording functionality of option.

        @Note: types showing strategy is last use first show, so we need to
        update the order each time.
        """
        cls.obj['Type']['default'] = baserec.type
        cls.obj['Payment']['default'] = baserec.payment
        cls.obj['Currency']['default'] = baserec.currency
        cls.obj['Tag']['default'] = baserec.tag

        cls.obj['Type']['type_order'] = _updOrder(
            cls.obj['Type']['default'][0], cls.obj['Type']['type_order']
        )
        cls.obj['Payment']['types'] = _updOrder(
            cls.obj['Payment']['default'], cls.obj['Payment']['types']
        )
        cls.obj['Currency']['types'] = _updOrder(
            cls.obj['Currency']['default'], cls.obj['Currency']['types']
        )
        cls.obj['Tag']['types'] = _updOrder(
            cls.obj['Tag']['default'], cls.obj['Tag']['types']
        )
        _exportObj(cls.obj, cls.objfile)

    @classmethod
    def getProjectBuffer(cls):
        if not cls.bufobj.has_key('ProjectBuf'):
            cls.bufobj['ProjectBuf'] = {}
        return cls.bufobj['ProjectBuf']

    @classmethod
    def setProjectBuffer(cls, bufobj):
        cls.bufobj['ProjectBuf'] = bufobj
        _exportObj(cls.bufobj, cls.buffile)

    @classmethod
    def getRecordBuffer(cls):
        if not cls.bufobj.has_key('RecordBuf'):
            cls.bufobj['RecordBuf'] = []
        return cls.bufobj['RecordBuf']

    @classmethod
    def setRecordBuffer(cls, bufobj):
        cls.bufobj['RecordBuf'] = bufobj
        _exportObj(cls.bufobj, cls.buffile)



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.Projects
    ~~~~~~~~~~~~~~~~~~~

    Implements the project's object for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

from datetime import date, timedelta
from ProjectParse import consumingProjectParse
from utils import to_date


class Project(object):
    def __init__(self, project_type = None):
        """Note that status is described to numeric code, has following
        situations:
                111 - Default status that nothing happens.
                222 - Project purpose is accomplish.
                333 - Project period was past, new record will not register.
                444 - Project had error occurs can't to continue process.
                555 - Some situation have happened need to focus.
        """
        self.p_type = project_type
        self.p_status = 111

    def register(self):
        pass

    def report(self):
        """There is a infomation report of Expenses project that exist
        in all project subclass.
        Note that self.p_progress_info should be a iterable sequence.
        """
        return self.p_progress_info

    def export_to_dict(self):
        retval = {}
        for key in self.__dict__:
            # the second expr is judge whether the value of key is None
            if key == 'p_start_time':
                retval[key] = self.__dict__[key].isoformat()
            elif key == 'p_period':
                retval[key] = self.__dict__[key].days
            elif key.startswith('p_'):
                retval[key] = self.__dict__[key]
        return retval

    def import_from_dict(self, importDict):
        for key in importDict:
            if key == 'p_start_time':
                self.__dict__[key] = to_date(importDict[key])
            elif key == 'p_period':
                self.__dict__[key] = timedelta(importDict[key])
            elif key.startswith('p_'):
                self.__dict__[key] = importDict[key]


class ConsumingProject(Project):
    def __init__(self, Target=-1, Bdate=date(1970, 1, 1), Period=0, 
                   Type = '', Optcode = '', Recursion = False):
        self.p_target_amount = float(Target)
        self.p_start_time = Bdate    # date of project started
        self.p_period = timedelta(int(Period))
        self.p_typename = Type    # a type name of BaseRecord attribute
        self.p_optcode = Optcode
        self.p_recursion = Recursion

        self._init_handler()
        super(ConsumingProject, self).__init__(project_type='statistic')

    def register(self, rdate, baserec, newrec=None, act='add'):
        """Register infomations of the specific record in the project for
        add, delete and update.
        """
        # the meaning of those status code described in this module above.
        # Note that these four status may have differently procesing
        # procedure in following operation inside.
        if self.p_status not in (111, 222, 555):
            return
        
        try:
            self._register_check(rdate, baserec)
        except ValueError:
            return
        
        self._update(rdate, baserec, newrec, act)
        self._target_check()

    def _update(self, rdate, baserec, newrec, act):
        """update infomation of amount for the project
        """
        if act == 'add':
            self.p_accumt_amount += baserec.amount
            self.p_recs_amount += 1
        elif act == 'del':
            self.p_accumt_amount -= baserec.amount
            self.p_recs_amount -= 1
        elif act == 'udp':
            self.p_accumt_amount += (baserec.amount - newrec.amount)
        else:
            # do nothing
            pass

    def _register_check(self, rdate, baserec):
        if not self._date_checke(rdate):
            # rdate not in the period (state != 0), raise Error to return
            raise ValueError
        if not self._type_checke(baserec):
            # baserec type is not match self.optcode, raise Error to return
            raise ValueError

    def _date_checke(self, rdate):
        """Judge a few situations of date of record and handle it.
        """
        if date.today() >= self.p_start_time + self.p_period:
            # Today is out of period, handle all situation.
            if self.p_recursion:
                # Init the recursive situation.
                self._recursive_init()
            else:
                self._finish()
        
        if self.p_start_time <= rdate < self.p_start_time + self.p_period:
            # The date of record is in the period.
            return True
        else:
            # The date of record is not in the period.
            return False

    def _type_checke(self, baserec):
        """Check self.typename of baserec whether match the self.optcode.
        Note that self.typename and self.optcode may not was specified that
        meant any type of consuming could be record in this project, and
        this function just return True and do nothing.
        """
        if not self.p_typename or not self.p_optcode:
            return True

        if self.p_typename == 'type' and not isinstance(self.p_optcode, tuple):
            # BaseRecord.type is constructed to (maintype, subtype), is a
            # 2-tuple, this situation is just filter maintype of type.
            if getattr(baserec, self.p_typename)[0] == self.p_optcode:
                return True
            else:
                return False

        # general situations solving
        if getattr(baserec, self.p_typename) == self.p_optcode:
            return True
        else:
            return False

    def _target_check(self):
        if self.p_accumt_amount >= self.p_target_amount:
            if not self.p_accomplish_date:
                # accomplish_date not had record.
                self.p_accomplish_date = date.today().isoformat()
                self.p_status = 222
        else:
            if self.p_accomplish_date:
                # Accomplish_date had been recorded.
                self.p_accomplish_date = ''
                self.p_status = 111
        # always analyze the data
        self._analyze()

    def _analyze(self):
        """
        """
        data = self.export_to_dict()
        self.p_progress_info = consumingProjectParse(data)

    def _init_handler(self):
        self.p_accumt_amount = 0.0    # accumulated amount
        self.p_recs_amount = 0    # amount of records were registered in project
        self.p_accomplish_date = ''    # accomplishing date of project purpose

        self.p_info_history = {}
        self.p_progress_info = []
        
        self.p_statdict = dict(
            hasQuarter = False,
            hasHalf = False,
            hasThreeQuarters = False,
            DateOfQuarter = '',
            DateOfHalf = '',
            DateOfThreeQuarters = ''
        )

    def _recursive_init(self):
        self.p_status = 111
        self.p_info_history[self.p_start_time] = self.p_progress_info
        self.p_start_time = self.p_start_time + self.p_period
        self._init_handler()

    def _finish(self):
        self.p_status = 333



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.Record
    ~~~~~~~~~~~~~~~~~

    Implements the unit object of expense's record for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""


class BaseRecord(object):
    """Basic structure of single item use to store record.
    """
    def __init__(self, Amount, Type, Payment, Currency, Tag='', Note=''):
        self.amount = float(Amount)
        # Note: type structure is a 2-tuple, e.g, (type, subtype)
        self.type = unicode(Type[0].encode('utf-8'), 'utf-8'), \
                    unicode(Type[1].encode('utf-8'), 'utf-8')
        self.payment = unicode(Payment.encode('utf-8'), 'utf-8')
        self.currency = unicode(Currency.encode('utf-8'), 'utf-8')
        self.tag = unicode(Tag.encode('utf-8'), 'utf-8')
        self.comment = unicode(Note.encode('utf-8'), 'utf-8')

    def __eq__(self, brec):
        return self.type == brec.type and \
               self.amount == brec.amount and \
               self.payment == brec.payment and \
               self.currency == brec.currency and \
               self.tag == brec.tag and \
               self.comment == brec.comment

    def __ne__(self, brec):
        return not self == brec

    def copy(self, nrec):
        self.type = nrec.type
        self.amount = nrec.amount
        self.payment = nrec.payment
        self.currency = nrec.currency
        self.tag = nrec.tag
        self.comment = nrec.comment



#! /usr/bin/env python
# -*- coding:utf-8 -*-
"""
    pyExpenses.Expense
    ~~~~~~~~~~~~~~~~~~

    Implements the schedual object for pyExpenses.

    :copyright: (c) 2012 by Jason Lai.
    :license: BSD, see LICENSE for more details.
"""

import os
import sys
import time
from decimal import Decimal
from datetime import date

from RecManip import RecManip
from ConfigManip import Config
from utils import buf2record, record2buf
import Projects
import RecParser as RP
import ErrorHandle as EH


class Expense(object):
    def __init__(self):
        self.rec_m = RecManip()    # Records Manipulator
        self.projects = {}
        self.isSetup = False

    def __getattribute__(self, name):
        """If isSetup attribute was False, when get a attributes of
        Expense's instance will raise AttributeError except setUp,
        addRecord and isSetup three attributes.
        @Note: Should call setUp method to avoid AttributeError.
        """
        if not name.startswith('__') and \
        not name.startswith('_') and \
        name not in ('setUp', 'addRecord', 'isSetup', 'rec_m', 'projects') and \
        not self.isSetup:
            raise AttributeError

        return super(Expense, self).__getattribute__(name)

    def updatePassword(self, oldpwd, newpwd):
        """Update the password of pyExpenses
        """
        self.rec_m.updatePassword(oldpwd, newpwd)

    def cancelPassword(self, oldpwd):
        """Do not use password authentication for pyExpenses
        """
        self.rec_m.cancelPassword(oldpwd)

    def setUp(self, rmArgs={}):
        """It's the method setup the Expense's instance, this should be
        called after initialise finish. 
        """
        self.rec_m.setUp(**rmArgs)
        
        self.isSetup = True
        self._setUpProjects()
        self._setUpRecords()

    def addRecord(self, rdate, baserec):
        if self.isSetup:
            self.rec_m.addItem(rdate, baserec)
            self._checkProjects(rdate, baserec, act='add')
            Config.setDefault(baserec)
        else:
            buf = record2buf(rdate, baserec)
            recbuf = Config.getRecordBuffer()
            recbuf.append(buf)
            Config.setRecordBuffer(recbuf)

    def deleteRecord(self, rdate, baserec):
        self.rec_m.delItem(rdate, baserec)
        self._checkProjects(rdate, baserec, act='del')

    def updateRecord(self, rdate, baserec, updaterec):
        self.rec_m.updateItem(rdate, baserec, updaterec)
        self._checkProjects(rdate, baserec, updaterec, act='upd')

    def listProject(self, ptype=None):
        if not ptype:
            return self.projects.iteritems()
        else:
            return (
                (name, proj)
                    for name, proj in self.projects.iteritems()
                        if proj.p_type == ptype
            )
        
    def addProject(self, name, project):
        if self.projects.has_key(name):
            return False
        self.projects[name] = project
        return True

    def deleteProject(self, name):
        del self.projects[name]

    def allRecords(self):
        """return a dataflow containing all records in storage.
        """
        return self.rec_m.getAll()

    def recordsInDaterange(self, bdate, edate):
        """return a dataflow containing specific records that specified
        by date range in storage.
        """
        return self.rec_m.date_range(bdate, edate)

    def figureOutRecords(self, bdate, edate, parsers = [], allrec=False):
        """
        """
        if allrec:
            data = self.allRecords()
        else:
            data = self.rec_m.date_range(bdate, edate)
        parser = RP.MainParser(data)
        parser.extend(parsers)
        return parser.parse()
    
    def saveProjects(self):
        savebuf = {}
        for name, proj in self.projects.items():
            savebuf[name] = proj.export_to_dict()
            savebuf[name]['projtype'] = proj.__class__.__name__
        Config.setProjectBuffer(savebuf)

    def saveRecords(self):
        self.rec_m.save()

    def save(self):
        self.saveProjects()
        self.saveRecords()
                
    def importData(self, filename):
        """import a backup file that was implemented by CSV format from 
        specify file, note that it will override original data."""
        self.rec_m.importRecord(filename)

    def exportData(self, filename):
        """Export textual data records that is implemented by CSV format
        into the specified file.
        """
        self.rec_m.exportRecord(filename)

    def resetAll(self):
        """This method reset all the data of the Expense to default status.
        Be careful when call this method, it should be make a warning
        message to the user.
        """
        self.rec_m.clear()
        for key in self.projects:
            del self.projects[key]
        Config.setProjectBuffer(dict())

    def _setUpProjects(self):
        projbuf = Config.getProjectBuffer()
        if projbuf:
            # projbuf is not a empty dict
            for name, dictbuf in projbuf.items():
                projtype = dictbuf['projtype']
                proj = Projects.__dict__[projtype]()
                proj.import_from_dict(dictbuf)
                self.projects[name] = proj

    def _setUpRecords(self):
        recbuf = Config.getRecordBuffer()
        if recbuf:
            # recbuf is not a empty list
            for buf in recbuf:
                rdate, rec = buf2record(buf)
                self.addRecord(rdate, rec)
        # set record buffer up to default value a empty list.
        Config.setRecordBuffer(list())

    def _checkProjects(self, rdate, baserec, newrec=None, act='add'):
        if not self.projects:
            # there is no project in self.projects
            return
        
        for proj in self.projects.values():
            if proj.p_type == 'statistic':
                proj.register(rdate, baserec, newrec, act)



#!/usr/bin/env python

import os
from setuptools import setup
from pyExpenses import __version__


setup(name='pyExpenses',
      version=__version__,
      description='pyExpenses is a python package for simply personal financial management.',
      author='Jason Lai',
      author_email='jasonlai256@gmail.com',
      maintainer='Jason Lai',
      maintainer_email='jasonlai256@gmail.com',
      url='https://github.com/JasonLai256/pyExpenses.git',
      packages=['pyExpenses'],
      license="BSD",
      platforms=["any"],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Other Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Office/Business :: Financial',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
 )
